autosave
' SB7GHZ
' Stellar Battle in the Seven Green Hills Zone by Vegipete 2022
' PicoMite VGA version - recommend version 5.07.05b15 or later
'   changes for PicoMiteVGA:
'   -Mode 2 - 320x240 resolution, 16 colours
'   -Lots of page/framebuffer changes
'      needs recent firmware for 3D Engine and FRAMEBUFFERS
'   -Atari (switched) joystick in PicoGAME-Mini (Port A of PicoGAME)
'      1   GP0   Up
'      2   GP1   Down
'      3   GP2   Left
'      4   GP3   Right
'      5   GP26  n.c.
'      6   GP14  Fire / Trigger
'      7   3V3   3V3
'      8   GND   GND
'      9   GP27  n.c.
'      Joystick processing routine is at very end of program
'   -Keys pressed on the console (arrows & SPACE) work, but not well.

'option angle degrees - not valid on PicoMite!
option default integer

' configure pins used for joystick
SetPin gp0,din
SetPin gp1,din
SetPin gp2,din
SetPin gp3,din
SetPin gp14,din

mode 2:cls  ' 320x240x16 colour
font 7  ' 6x8
const TANKSPEED=15
const SHOTSPEED=150
const ARENASIZE=6E4
const BLASTRADIUS=5E4
const MMHQ = MM.HRES/4
const ASSETCOUNT=20

' object array holds all the items in the arena:
'  0) player-unused
'  1) player shot 1
'  2) player shot 2
'  3) enemy shot 1
'  4) enemy shot 2
'  5) fuel bay
'  6) stargate
'  7) enemy 1
'  8) enemy 2
'  9-20) obstacles
'  -------
'  0) exists
'  1) x position
'  2) y position
'  3) angle
'  4) velocity    (shot: dx)
'  5) (shot velocity: dy)
'  6) type: 1=shot, 2=pyramid, 3=cube, 4=fuel, 5=stargate, 6=tank1, 7=tank2
dim object(ASSETCOUNT,6)
dim distaway(ASSETCOUNT),distindx(ASSETCOUNT) ' for drawing objects far to near
dim radar(ASSETCOUNT,2)   ' blips on radar: x,y,colour

dim ctable(8)=(&hA52A2A,65280,33023,32768,16384,0,&h80FF,&hFF00C0,&h404040)
dim hills(18)=(0,3,7,2,9,3,6,2,8,12,9,1,6,2,9,3,6,1,0) ' fold this into 3D data

dim fv(64),fcnt(20),ecol(20),fcol(20)
dim s(200)    ' longstring for holding 3D object data for decoding
dim t(9)      ' tank AI counters
dim float head,fuel,shield,v(2,32),ct,st,q(4)

' 3D object data - carefully compressed
' Two big compression techniques:
' 1) numbers compressed into ascii codes, where "A"=0 (no commas, single byte per value)
' 2) entire repeated data array can be skipped
'    For example, the pyramid obstacles have chopped off tops, so data-wise, they are
'    identical to the cubes. Only the the actual vertex coordinates are unique between
'    the two object. All the other arrays can be rpresented by a single 'repeat' byte.
LA "EEDDDDHAAAHAAA>CAAHADCAAEJJLABCCBDBADACDIFEEEEEEEEEEDDDDD8A8JA8JAJ8AJ@S@BS@B"
LA "SB@SBNTAEFBBFGCCGHDDHEAHGFEIF!!!=A=EA=EAE=AE=I=EI=EIE=IEZANKEEEEEEDDDDEEEEEE"
LA "BBBBAAAAACBBBB<A<FA<FAF<AF?B?CB?CBC?BC>G>DG>DGD>GDAIAUdAEFBBFGCCGHDDHEAHGFEI"
LA "JKLIMJJMKKMLLMINK!EEEEEEGGGGGGGGBBDDDD@A@BA@BAB@AB?M?CM?CMC?MCAM7GMAAMK;MAAQ"
LA "AKALIDDDDEEEDEEEEEEEEBBBBCCCEDA9DAI>AI>A9AF:BD@BEJ@EJ@D@AE=AFJU[AEBBECCEDADE"
LA "FGHIFJKGIHKJGKHPNEEEEEDDDDEEEDEEEEEEEEEEEEEEDDDDCCCCBBBEDA9DAJ>AJ>A9ED8EDK=D"
LA "K=D8AH;BFCBGK@GK@FCAG?AHKPoABCDAEFBBFGCCGHDADHEEIFFIGGIHHIEJKLMJNOKMLONKOL  "
k=1
do
  nv=NB() ' read number of vertices for next object
  if nv<-32 then exit do
  nf=NB() ' read number of faces
  RC(fcnt())  ' read number of vertices for each face
  RC(ecol())  ' read colour of edge of each face
  RC(fcol())  ' read colour of each face
  for j=0to nv-1:v(0,j)=NB():v(1,j)=NB():v(2,j)=NB():next ' read vertex coordinates
  sc=NB():math scale v(),sc,v() ' read and scale vertex cloud
  tot=NB()  ' total number of vertices of faces, skip if 0 (no change)
  if tot then for j=0to tot:fv(j)=NB():next  ' vertices for each face
  draw3d create k,nv,nf,1,v(),fcnt(),fv(),ctable(),ecol(),fcol()
  inc k
loop

for i=1to 8:read object(i,6):next:data 1,1,1,1,4,5,6,7 ' object types

framebuffer create 'PW 3
framebuffer write F
cls

' paint ringed planet sprite - well, no, not yet.
'circle MM.HRES/2,MM.VRES/2,11,,4,ctable(2),ctable(2)    ' rings OD
'circle MM.HRES/2,MM.VRES/2, 7,,4,0,0                    ' rings ID
'blit read 8,MM.HRES/2-30,MM.VRES/2,60,30                   ' store rings in front of planet
'circle MM.HRES/2,MM.VRES/2,30,,,ctable(3),ctable(3)     ' draw planet
'blit write 8,MM.HRES/2-30,MM.VRES/2,4                      ' restore rings in front of planet
''image rotate 150,100,100,100,150,100,15     ' rotated looks cooler
'blit read 1,315,268,170,64                  ' save ringed planet image for later use

' Start a new game
head=0       ' player heading
fuel=99
shield=99
level=1
demo=0    ' demo mode on to start
'wpage=0   ' page flipping
newpress=1

' bigger zoom=narrower field of view : zoom in
' 500 is approx 90 degrees (on CMM2, mode 1)
zoom=200    ' approx 90 degrees on PicoMiteVGA mode 2
draw3d camera 1,zoom,0,75

MakeObstacles

settick 25,MM,1   ' start timer interrupt for sound and shield repair

do
  ' test for key presses
  k = asc(inkey$)
  if k then ServiceKey k    ' found a key in the console buffer?
  
  ReadJoystick
  
  ' MODIFIER pressed-create a new shot of there is none and hasn't been one for a while
  if shotpress then
    if newpress then
      newpress=0  ' edge select the keypress so only one shot is created
      for i=1to demo*2
        if object(i,0)=0 then v2=25:NewShot(i,0,0,-head):i=3   ' force loop to end
      next
    endif
    shotpress = 0
  else
    newpress=1
  endif
  
  ' move various objects
  ' 1) move shots, test for hits
  for i=1to 4
    if object(i,0)then  ' shot exists
      inc object(i,0),-sgn(object(i,0))  ' shot expires after a while
      target = MoveBlocked(i,7,object(i,4),object(i,5))  ' shot hits something
      if target>6 then  ' hit an enemy tank or an obstacle
        object(i,0)=-object(i,0)          ' shot gone
        if target<9 then                  ' hit an enemy tank
          v1=75                           ' play explosion sound
          object(target,1)=ARENASIZE/2    ' new tank x-coord
          object(target,2)=ARENASIZE/target   ' new tank y-coord
          object(target,3)=rnd*360        ' new tank angle
          inc score,demo                  ' no score increase in demo mode
          if score>level*9 and object(6,0)=0 then v4=99:object(6,0)=1   ' stargate appears
        endif
      endif
    endif
  next
  ' 2) stargate slowly turns
  inc object(6,3),1
  ' 3) enemies turn towards player and move
  TurnEnemy(7)
  TurnEnemy(8)
  ' 4) slowly turn view if in demo mode
  if demo=0 then head=head+1/5
  ' 5) keep heading in range (0,360]
  if head>360 then head=head-360   ' can't use MOD because we loose the decimal part
  if head<0 then head=head+360
  
  ' Test for game over
  if shield<0 then
    framebuffer write N  ' PW 0
    for i = 0 to 999
      line 0,600*rnd,800,600*rnd,,ctable(0)
      pause 2
    next
    text MM.HRES/2,MM.VRES/2,"GAME OVER",CT,3,,ctable(0)
    pause 4000
    demo=0  ' back to demo mode
    shield=99
    NewLevel
  endif

  ' draw scene
  framebuffer write F
  DrawScene
  if qx+qy then sprite scroll qx,qy:qx=0:qy=0   ' shake screen if hit

  ' wait for framerate, then display
  do : loop until timer>25
  timer=0
  framebuffer copy F,N,B  ' copy frame buffer to display on next VBL
loop

end

'==================
sub ServiceKey(k)
  ' space bar to end demo mode
  if k+demo=32 then head=0:level=1:demo=1:score=0:NewLevel:end sub
  
  ' SPACE BAR = new shot
  shotpress = (k=32)
  
  ' turn left or right if not demo mode
  head = (head+360+((k=131)-(k=130))*demo)mod 360 
  
  ' move forward or reverse
  MoveForward 50*((k=128)-(k=129))
end sub

'==================
' Since the camera can't move, everything else must move instead
' Move forward requested distance in direction we are facing,
' unless blocked by items 7 thru 20.
' Handle encounters with fuel bay and star gate
sub MoveForward(dist)

  if fuel<0 or dist*demo=0 then exit sub   ' no gas left, no distance or demo mode
  fuel=fuel-(fuel>0)/200   ' burn more fuel while moving
  dx=sin(rad(head))*dist
  dy=cos(rad(head))*dist
  
  i = MoveBlocked(0,5,dx,dy)  ' player try to move forward
  if i<6 then  ' move forward and clamp into the arena
    for i=1to ASSETCOUNT:Clamp2Arena(object(i,1),-dx):Clamp2Arena(object(i,2),-dy):next
  elseif i=6 then  ' drive thru star gate?
    framebuffer write N ' draw stargate effect direct on main screen
    for i=1to 999
      n=rnd*360
      line MM.HRES/2,MM.VRES/2,MM.HRES/2+500*sin(rad(n)),MM.VRES/2+500*cos(rad(n)),,&hFFFF
      pause 1
    next
    head=(head+180)mod 360  ' turn 180 degrees to start next level
    inc level
    NewLevel
  else
    v3=-2             ' bump sound
    shield=shield-1   ' damage from crashing into something
  endif
end sub

'==================
' create a new shot
sub NewShot(n,x,y,h)
  object(n,0)=90*demo  ' shot duration
  object(n,1)=x
  object(n,2)=y
  object(n,3)=h
  object(n,4)=-sin(rad(h))*SHOTSPEED  ' dx
  object(n,5)= cos(rad(h))*SHOTSPEED  ' dy
end sub

'==================
sub Clamp2Arena(c,d):c=(c+d+1.5*ARENASIZE)mod ARENASIZE-ARENASIZE/2:end sub

'==================
' Can object n move dx,dy?  n is 0,1,2,3,4,7,8
' Test against selectable other objects  (, except fuel bay)
function MoveBlocked(n,f,dx,dy)
  'local x,y,j
  MoveBlocked=1
  x=object(n,1):Clamp2Arena(x,dx)
  y=object(n,2):Clamp2Arena(y,dy)
  
  ' enemy tank can't drive through player
  if n>6 and x^2+y^2<BLASTRADIUS then exit function
  
  for j=f to 20
    j=j+(j=n)+(n+4=j)  ' tank can't block or shoot itself
    if (object(j,1)-x)^2+(object(j,2)-y)^2<BLASTRADIUS then MoveBlocked=j:exit function
  next
  MoveBlocked=0
  if n then object(n,1)=x:object(n,2)=y ' update position if not blocked
end function

'==================
' Create and display a new level
sub NewLevel
  MakeObstacles
  framebuffer close     ' close the framebuffer (not enough memory for both)
  framebuffer layer     ' create the overlay layer
  framebuffer copy N,L  ' copy whatever is currently visible to the top layer
  framebuffer write N   ' draw on the main screen
  DrawScene
  ' open iris new world over star travel spray
  framebuffer write L   ' erase star travel spray which is on overlay
  for i =1 to MM.HRES/2
    box MM.HRES/2-i,MM.VRES/2-i,i*2,i*2,3,&hFFFF,0
    pause 200/i
  next
  framebuffer close     ' done with overlay layer
  framebuffer create    ' recreate the framebuffer
  framebuffer write L
end sub

'==================
sub TurnEnemy(n)
  local h,te
  
  ' busy with obstacle avoidance turn?
  if t(n)then inc t(n),-sgn(t(n)):object(n,3)=(object(n,3)+sgn(t(n))*2)mod 360:exit sub
  ' adjust angle towards origin
  te = deg(atan2(object(n,1),-object(n,2)))
  inc te,360*(te<0)  ' direction toward origin/player
  h = (object(n,3)-te)mod 360
  
  ' shoot if aimed at origin/player and no shot in flight
  if h=0 and object(n-4,0)=0 then NewShot(n-4,object(n,1),object(n,2),object(n,3))
  te=-sgn(h)
  if abs(h)>180 then te=-te 'NG te
  object(n,3)=(object(n,3)+te)mod 360  ' new enemy heading
  
  ' enemy try to move forward
  if MoveBlocked(n,5,-sin(rad(object(n,3)))*object(n,4),cos(rad(object(n,3)))*object(n,4))then
    ' can't move that way, so start a random turn left or right
    t(n)=60*sgn(rnd-.5)
  endif
end sub

'==================
' Create a bunch of objects
sub MakeObstacles
  'local i

  for i=1to 4:object(i,0)=0:PositionItem(i+4):next  ' clear shots, create FB,SG, enemy tanks
  
  object(7,4)=TANKSPEED*2 + 2*level  ' speedy!
  object(8,4)=TANKSPEED + 2*level
  
  for i=9to ASSETCOUNT
    PositionItem(i)
    object(i,6)=2 + (rnd<.5)  ' assume it's a pyramid to start, randomly change to cube
  next
end sub

'==================
sub PositionItem(n)
  local  x,y

  object(n,0)=(n<>6)    ' items visible except star gate not visible at start
  do
    x=rnd*ARENASIZE-ARENASIZE/2
    y=rnd*ARENASIZE-ARENASIZE/2
  loop while x^2 + y^2<4E6
  object(n,1)=x  ' x position
  object(n,2)=y  ' y position
  object(n,3)=rnd*360  ' angle

end sub

'==================
' test player has been hit by enemy shot n
sub PlayerHit(n)
  if object(n,0)>0 then
    object(n,0)=-object(n,0)
    inc shield,-5*demo
    qx=rnd*10-5 ' wiggle the screen image
    qy=rnd*10-5

    v3=-3
  endif
end sub

'==================
' Draw background and all the visible 3D objects, from farthest to nearest
' image gets drawn on currently active layer
sub DrawScene
  local i,n,ix,j

  ' calculate distance to each item
  for i=1to ASSETCOUNT:distaway(i)=object(i,1)^2+object(i,2)^2:next
  
  ' test if player hit by enemy shot
  for i=3to 4:if distaway(i)<BLASTRADIUS then PlayerHit(i)
  next

  fuel=fuel-demo*(fuel>0)/200  ' use up fuel
  ' sitting in fuel bay?
  if distaway(5)<BLASTRADIUS then fuel=100:shield=min(100,shield+sr/500):sr=0

  sort distaway(),distindx(),1  ' sort farthest to nearest (descending distance)

  ct=cos(rad(head))
  st=sin(rad(head))

  'framebuffer write F   ' PW wpage+1
  cls
  line 0,MM.VRES/2,MM.HRES,MM.VRES/2,1,ctable(3)  ' horizon line

  'if head>239 and head<340 then blit write 1,3100-head*9.6,MM.VRES/10  ' Saturn
  circle MM.HRES-head*3.84,MM.VRES/4,10,0,1,,ctable(3)  ' moon
  circle MM.HRES-head*3.84-3,MM.VRES/4+1,10,0,1,,0      ' change it to a crescent
  circle MM.HRES*2-head*3.84,MM.VRES/10,15,0,1,,rgb(128,255,0) ' sun

  ' draw those dear green hills
  ix=(head-.49)\20       ' index at center (0.49 for rounding reasons)
  n=-head*4 mod MMHQ     ' = MM.HRES/4
  for i=0to 5
    j=(ix+i+15)mod 18
    line n,MM.VRES/2-hills(j)*5,n+MMHQ,MM.VRES/2-hills(j+1)*5,,ctable(3)
    n=n+MMHQ
  next

  math set 0,radar()  ' clear all blips from radar screen
  for i=1to ASSETCOUNT
    n=distindx(i)
    
    if object(n,0)>0 then  ' exists/visible
    
      ' revolve items
      x=object(n,1)*ct - object(n,2)*st
      y=object(n,1)*st + object(n,2)*ct
      
      ' save blip to radar image data
      radar(i,0) = x * 70 / ARENASIZE
      radar(i,1) = y * 70 / ARENASIZE
      radar(i,2) = 6*(object(n,6)=5)+(object(n,6)=4)  ' stargate in blue, fuel bay in green
      
      ' draw the 3D object
      ' min 220 to avoid drawing bug?
      if y>160 and abs(x)<y then
        math q_euler rad(object(n,3)+(head)),0,0,q()
        draw3d rotate q(),object(n,6)
        draw3d write object(n,6),x,0,y
      endif
    endif
  next

  ' draw the radar image
  box MM.HRES/2-12,MM.VRES-54,53,53,1,rgb(grey),0 'redraw frame
  for i=1to ASSETCOUNT  ' draw radar image
    if abs(radar(i,0)) < 26 and abs(radar(i,1)) < 26 then
      pixel MM.HRES/2+14+radar(i,0),MM.VRES-28-radar(i,1),ctable(radar(i,2))
    endif
  next
  pixel MM.HRES/2+14,MM.VRES-28,rgb(yellow)   ' player in center
  
  ' draw headings and stuff
  text MM.HRES/2,6,"LEVEL:"+str$(level,2)+"  SCORE:"+str$(score,3),CT,,,ctable(1),-1
  text MM.HRES/2,MM.VRES-55,"S F  HEADING: "+str$(head,3,0),CB,,,ctable(1),-1
  DrawGauge(MM.HRES/2-54,shield)
  DrawGauge(MM.HRES/2-42,fuel)
  if demo=0 then text MM.HRES/2,MM.VRES/4,"[FIRE] to Play",CB,3,,ctable(1),-1
  
  ' erase effects of drawing routine bug
  line 0,0,0,MM.VRES,1,0
  line MM.HRES-1,0,MM.HRES-1,MM.VRES,1,0
end sub

' Draw gauge to show shield or fuel
sub DrawGauge(x,g)
  box x,MM.VRES-53,10,52,1,&h404040,0
  if g>0 then box x+1,MM.VRES-3-g/2,8,g/2,4,ctable(g>25)
end sub

' Read an array of data, or leave previous untouched
sub RC(c())
  local j,tmp
  tmp=NB():if tmp<>-32 then c(0)=tmp:for j=1to nf-1:c(j)=NB():next
end sub

' Read next byte from data string
function NB()
  static n:NB=lgetbyte(s(),n)-65:inc n
end function

' Build lonstring of data
sub LA z$
  longstring append s(),z$
end sub

' Timer interrupt
sub MM  'fade out sound
' Sorry, no sound for now because noise type isn't supported
  'play sound 1,b,n,99,v1/3
  v1=v1-(v1>0)  ' decrease volume of explosion sound
  'play sound 2,b,p,99-v2,v2
  v2=v2-(v2>0)  ' decrease volume of shot sound
  'play sound 3,b,n,200,20*(v3<>0)
  v3=v3+(v3<0)
  'play sound 4,b,t,499-99*sin(rad(v4)),v4/4
  v4=v4-(v4>0)  ' star gate appearing
  inc sr  ' shield repair
end sub
'=============

' Read and handle joystick inputs
' Modify to suit your hardware
' Joystick commands are converted to equivalent keystrokes
sub ReadJoystick
  if not Pin(gp0)  then   ' forward     all are active low
    ServiceKey(128)
  endif
  if not Pin(gp1)  then   ' backward    0 - do something
    ServiceKey(129)
  endif
  if not Pin(gp2)  then   ' turn left   1 - do nothing
    ServiceKey(130)
  endif
  if not Pin(gp3)  then   ' turn right
    ServiceKey(131)
  endif
  if not Pin(gp14) then   ' fire
    if demo then  ' in game mode so try create a new shot
      shotpress=1
    else  ' in demo mode, so start a new game
      head=0:level=1:demo=1:score=0:NewLevel
    endif
  endif

end sub
